"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import ctypes
import sdl2.sdlmixer as sdlmixer
import weakref

class Music:
	def __init__(self, filename):
		self._music = sdlmixer.Mix_LoadMUS(filename)
		if not self._music:
			raise Exception("Could not load music file: "+filename)


	def play(self, fade_in_ms = 0, volume=None):
		if fade_in_ms <= 0:
			sdlmixer.Mix_PlayMusic(self._music, -1)
		else:
			sdlmixer.Mix_FadeInMusic(self._music, -1, int(fade_in_ms))

		if volume is not None:
			self.setVolume(volume)


	def stop(self, fade_out_ms = 0, on_stop=None):
		fade_out_ms = 0 #FIXME: make fade out work
		if on_stop:
			#CFUNC = ctypes.CFUNCTYPE(None) # FIXME: see why this fails on winbugs
			#sdlmixer.Mix_HookMusicFinished(CFUNC(on_stop))
			pass
		if fade_out_ms <= 0:
			sdlmixer.Mix_HaltMusic()
		else:
			sdlmixer.Mix_FadeOutMusic(int(fade_out_ms))

	def setVolume(self, volume): # in range 0..1
		sdlmixer.Mix_VolumeMusic(int(max(min(128.0*volume+0.5,128),0)))

	def isPlaying(self):
		return sdlmixer.Mix_PlayingMusic()

	def __del__(self):
		sdlmixer.Mix_FreeMusic(self._music)

_sound_channels = {}


class SoundInstance:
	def __init__(self, sound, channel, on_stop, volume):
		self._sound = sound
		self._channel = channel
		self._playing = sdlmixer.Mix_Playing(self._channel)

		if self._playing:
			_sound_channels[self._channel] = weakref.ref(self)
			self._on_stop = on_stop
		else:
			_sound_channels[self._channel] = None
			if on_stop is not None:
				on_stop(self)
			self._on_stop = None

		if volume:
			self.setVolume(volume)


	@property
	def sound(self):
		return self._sound

	def stop(self, fade_out_ms=0, on_stop=None):
		if not self._playing:
			if on_stop is not None:
				on_stop(self)
			return

		if on_stop is not None:	self._on_stop = on_stop

		if fade_out_ms<=0:
			sdlmixer.Mix_HaltChannel(self._channel)
		else:
			sdlmixer.Mix_FadeOutChannel(self._channel, int(fade_out_ms))


	def setVolume(self, volume): # in range 0..1
		if not self._playing: return
		sdlmixer.Mix_Volume(self._channel, int(max(min(128.0*volume+0.5,128),0)))


	def isPlaying(self):
		return self._playing


	def _onStop(self):
		self._playing = False
		if self._on_stop:
			self._on_stop(self)
		self._on_stop = None
		self._channel = None


class Sound:
	def __init__(self, filename):
		self._sound = sdlmixer.Mix_LoadWAV(filename)
		if not self._sound:
			raise Exception("Could not load audio file: "+filename+ "  " + sdlmixer.Mix_GetError())


	def setVolume(self, volume): # in range 0..1
		sdlmixer.Mix_VolumeChunk(self._sound, int(max(min(128.0*volume+0.5,128),0)))

	# loops can be:
	#  an int for the number of times the audio is played
	#  False, same as 1
	#  True, for infinite looping
	# max_time is in seconds
	def play(self, loops=False, max_ms=None, volume=None, fade_in_ms=0, on_stop=None):
		if loops is True: loops = -1
		elif loops is False: loops = 0
		if max_ms is None: max_ms = -1
		else: max_ms = int(max_ms * 1000)

		if fade_in_ms > 0:
			channel = sdlmixer.Mix_FadeInChannelTimed(-1, self._sound,loops,int(fade_in_ms),max_ms)
		else:
			channel = sdlmixer.Mix_PlayChannelTimed(-1, self._sound,loops,max_ms)

		#return SoundInstance(self, channel, on_stop, volume)


	def __del__(self):
		sdlmixer.Mix_FreeChunk(self._sound)


def _onChannelStops(channel):
	global _sound_channels
	instance = _sound_channels[channel]
	if instance:
		_sound_channels[channel] = None
		instance = instance()
		if instance:
			instance.onStop()


def init():
	global _sound_channels
	sdlmixer.Mix_AllocateChannels(10)
	#CFUNC = ctypes.CFUNCTYPE(None, ctypes.c_int)
	#sdlmixer.Mix_ChannelFinished(CFUNC(_onChannelStops))
