"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as N
import linalg
from packedarr import PackedArray


class BezierGroup(PackedArray):
	def __init__(self, degree=3, dimensions=2, initial_size=100,max_steps=20):
		PackedArray.__init__(self, initial_size)
		self._degree = degree
		self._dimensions = dimensions
		self._id2loc= {}
		self._loc2id = []
		self._max_id = 0
		self._max_steps = max_steps
		self._computed_steps = 0
		self._dirty_from = initial_size
		self._dirty_to = 0

		self._p = N.zeros([initial_size, degree+1,dimensions], dtype=N.float)
		self._q = N.zeros([initial_size, max_steps + degree+1, dimensions], dtype=N.float)


	def _resize(self, new_size):
		self._p = N.resize(self._p, [new_size, self._degree+1,self._dimensions])
		self._q = N.resize(self._q, [new_size, self._max_steps+self._degree+1, self._dimensions])


	def _moveData(self, to_loc, from_loc):
		self._p[to_loc] = self._p[from_loc]
		self._q[to_loc] = self._q[from_loc]


	def _getGetters(self):
		def val_getter(loc):
			return self._q[loc]
		def par_getter(loc):
			return self._p[loc]
		return val_getter, par_getter

	@property
	def dimensions(self):
		return self._dimensions

	def addCurve(self, points, recompute = True):
		loc, id = PackedArray._add(self)

		self._dirty_to = max(self._dirty_to,loc)
		self._dirty_from = min(self._dirty_from,loc)

		self._p[loc] = points

		if recompute: self.compute()

		return id


	def removeCurve(self, id, recompute = True):
		size = PackedArray._remove(self,id)

		self._dirty_to = min(self._dirty_to,size)

		if recompute: self.compute()


	def moveCurve(self, id, points):
		loc = self.loc(id)
		self._p[loc] = points

		self._dirty_to = max(self._dirty_to,loc)
		self._dirty_from = min(self._dirty_from,loc)


	def compute(self, t0=0, t1=1.0, steps=0, force = False):
		c = self.size
		if c == 0: return

		if force:
			cfrom = 0
			cto = c
		else:
			cfrom = self._dirty_from
			cto = self._dirty_to+1
			if cfrom >= cto: return

		n = self._degree

		if steps == 0:
			steps = self._max_steps

		if steps > 1:
			dt = float(t1-t0) / float(steps-1)
		else:
			dt = 0

		p = self._p[cfrom:cto]
		q = self._q[cfrom:cto]

		for s in xrange(steps):
			t = t0 + s * dt

			qv = q[:,s:s+n+1]
			qv[:,:,:] = p[:,:,:]

			t1 = 1.0-t

			for k in xrange(n):
				for i in xrange(0,n-k):
					qv[:,i] = t1 * qv[:,i] + t * qv[:,i+1]

		self._dirty_from = self._size
		self._dirty_to = 0

		self._computed_steps = steps


	def get(self, curve_id, pt_num):
		return self._q[self.loc(curve_id), pt_num]


	def getCurve(self, curve_id):
		return self._q[self.loc(curve_id), 0:self._computed_steps]


	def length(self, curve_id):
		loc = self.loc(curve_id)

		def reduceFn(prev, next):
			if prev is None:
				return (0.0,next)
			else:
				d = linalg.norm(prev[1] - next)
				return (prev[0]+d, next)

		return reduce(reduceFn, self._q[loc, 0:self._computed_steps], None)[0]


	def iter(self):
		return self._q[:self.size]


class Spline:
	def __init__(self, degree=3, dimensions=2,max_steps=20):
		self._degree = degree
		self._dimensions = dimensions
		self._max_steps = max_steps





if __name__ == "__main__":
	print("test...")

	bez = BezierGroup(3,2)

	bez.addCurve([(0,0),(0,1),(1,1),(1,0)])
	bez.addCurve([(0,0),(5,0),(0,5),(5,5)])

	bez.compute()

	import matplotlib.pyplot as plt

	c1 = bez.getCurve(0)
	c2 = bez.getCurve(1)

	plt.plot(c1[:,0],c1[:,1], 'r', c2[:,0],c2[:,1], 'g')
	plt.show()



