"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from resources import R
from clickable import Clickable
from entity import Entity
from sprite import StaticSprite
from sdl2 import *
import numpy as N

class Button(StaticSprite):
	def __init__(self, scene, src, pos, scale=1, on_click_cb=None, cb_data = None, background = None, min_size = None, padding = 5, bg_tint = (1,1,1), z = 0):
		if isinstance(src, basestring):
			self._foreground= R.texture(src)
		else:
			self._foreground = src

		bg_tint = N.array(bg_tint, dtype=N.float)

		if isinstance(background, basestring):
			self._background = R.texture(background, blend_mode = SDL_BLENDMODE_ADD)
		else:
			self._background = background


		self._hover_tint = (0.9,0.5,0.5)
		self._pressed_tint = (0.5,0.5,0.5)
		self._regular_tint = (0.9,0.9,0.9)

		self._bg_hover_tint = tuple(N.array(self._hover_tint) * bg_tint)
		self._bg_pressed_tint = tuple(N.array(self._pressed_tint) * bg_tint)
		self._bg_regular_tint = tuple(N.array(self._regular_tint) * bg_tint)

		self._hover_sound = R.sound("blip.wav")
		self._click_sound = R.sound("click.wav")


		self._tint = self._regular_tint
		self._bg_tint = self._bg_regular_tint

		self._padding = padding
		self._min_size = min_size or (1,1)

		self._bg_rect = SDL_Rect(0,0,1,1)

		StaticSprite.__init__(self, scene, self._foreground, pos, scale, z=z)


		self._evt_handler = Clickable(self, self)

		self._on_click_cb = on_click_cb
		self._cb_data = cb_data

		self.recalcRects()

	def recalcRects(self):
		w = max(self._rect.w + self._padding * 2, self._min_size[0])
		h = max(self._rect.h + self._padding * 2, self._min_size[1])

		self._x_ofs = float (w - self._rect.w) * 0.5
		self._y_ofs = float (h - self._rect.h) * 0.5

		self._bg_rect.x = int(self._rect.x - self._x_ofs + 0.5)
		self._bg_rect.y = int(self._rect.y - self._y_ofs + 0.5)
		self._bg_rect.w = int(w + 0.5)
		self._bg_rect.h = int(h + 0.5)


	def moveTo(self, pos): #FIXME: only works when centered = True
		StaticSprite.moveTo(self,pos)
		self.recalcRects()


	def entityType(self):
		return Entity.BUTTON_TYPE

	def categories(self):
		from scene import Scene
		return (Scene.RENDERABLE, Scene.LISTENER, Scene.STATIC)

	def processEvent(self, evt):
		return self._evt_handler(evt)

	def isPointOver(self, x, y):
		r = self._bg_rect
		dx = x - r.x
		dy = y - r.y
		# bounding box check
		if dx > r.w or dx < 0 or dy > r.h or dy < 0:
			return False
		return True

	def paint(self):
		if self._background:
			self._background.setTint(self._bg_tint)
			self._background.paintRect(self._bg_rect)
		self._foreground.setTint(self._tint)
		self._foreground.paintRect(self._rect)


	def _updateTextures(self):
		if self._evt_handler.pressed:
			self._tint = self._pressed_tint
			self._bg_tint = self._bg_pressed_tint
		elif self._evt_handler.hover:
			self._tint = self._hover_tint
			self._bg_tint = self._bg_hover_tint
		else:
			self._tint = self._regular_tint
			self._bg_tint = self._bg_regular_tint

	def _onPress(self, *args, **kargs):
		self._updateTextures()

	def _onRelease(self, *args, **kargs):
		self._updateTextures()

	def _onHover(self, *args, **kargs):
		self._updateTextures()
		self._hover_sound.play()

	def _onBlur(self, *args, **kargs):
		self._updateTextures()

	def _onClick(self,who,x,y):
		self._click_sound.play()
		if self._on_click_cb:
			self._on_click_cb(self, self._cb_data)

	def __str__(self):
		return "[Button %s]"%self._entity_id
