"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import re
import random
import numpy as N

class Level:

	assign_re = re.compile(r"^([*][^:= \t]*)\s*[:=]?\s*(.*?)\s*$")
	split_re = re.compile(r"(?:\s+)|(?:\s*,\s*)")

	def __init__(self, from_file=None):
		self.clear()
		if from_file:
			self.loadLevel(from_file)


	def clear(self):
		self._nodes = []
		self._links = []
		self._good_blob_node = 0
		self._bad_blob_node = 0
		self._background = None
		self._music = None
		self._enemy_time = 5
		self._user_time = 5
		self._seed = None
		self._total_links = 0
		self._total_nodes = 0

	@property
	def nodes(self):
		return self._nodes

	@property
	def links(self):
		return self._links

	@property
	def good_blob_start_node(self):
		return self._good_blob_node

	@property
	def bad_blob_start_node(self):
		return self._bad_blob_node

	@property
	def enemy_time(self):
		return self._enemy_time

	@property
	def user_time(self):
		return self._user_time

	@property
	def bg_image(self):
		return self._background

	@property
	def music(self):
		return self._music

	def _getAssignment(self, line):
		mo = self.assign_re.match(line)
		if not mo:
			print "Don't know how to parse: %s", line
			return None
		return mo.group(2)

	def _splitLine(self, line, max=0, cast_to=None):
		pieces = self.split_re.split(line, max)
		print "Line: ",line
		print "Pieces: ",pieces
		if not cast_to:
			return pieces
		else:
			return map(cast_to, pieces)

	def fullRandom(self):
		self.automatic(None, random.randint(1,5))
		self._background = "bg_r.jpg"


	def automatic(self, seed, difficulty):
		self.clear()

		print "Difficulty: ",difficulty

		self._enemy_time = 5 - max(difficulty*0.3, 2)
		self._user_time = 5 - max(difficulty*0.3, 2)


		if difficulty <= 4:
			n = 2 ** (difficulty-1) # 1,2,4,8
		else:
			n = difficulty*2 # 10,12,14,16

		total_nodes = n+4
		total_links = int((n+3) * 2)

		self._autogen(seed, total_nodes, total_links)
		self._good_blob_node = 0
		self._bad_blob_node = 1
		self.spreadNodes()

		#TODO: don't hardcode the number of backgrounds'
		self._background = "bg%s.jpg"%((difficulty % 6)+1)


	def loadLevel(self,from_file):
		self.clear()
		state = -1
		with open(from_file, "rt") as f:
			for line in f.readlines():
				line = line.strip()
				if len(line)==0 or line[0] == "#": continue

				if line[0] == "*":
					if line=="*nodes":
						state = 0
					elif line == "*links":
						state = 1
					elif line.startswith("*user_node"):
						a = self._getAssignment(line)
						if a: self._good_blob_node = int(a)
						state = -1
					elif line.startswith("*enemy_node"):
						a = self._getAssignment(line)
						if a: self._bad_blob_node = int(a)
						state = -1
					elif line.startswith("*music"):
						a = self._getAssignment(line)
						if a: self._music = a
						state = -1
					elif line.startswith("*background"):
						a = self._getAssignment(line)
						if a: self._background = a
						state = -1
					elif line.startswith("*user_time"):
						a = self._getAssignment(line)
						if a: self._user_time = float(a)
						state = -1
					elif line.startswith("*enemy_time"):
						a = self._getAssignment(line)
						if a: self._enemy_time = float(a)
						state = -1
					elif line.startswith("*seed"):
						a = self._getAssignment(line)
						if a: self._seed = int(a)
						state = -1
					elif line.startswith("*total_nodes"):
						a = self._getAssignment(line)
						if a: self._total_nodes = int(a)
						state = -1
					elif line.startswith("*total_links"):
						a = self._getAssignment(line)
						if a: self._total_links = int(a)
						state = -1
					else:
						print "Don't know how to parse: %s", line
						state = -1
					continue

				if state == 0: # read a node's coordinate
					x,y = self._splitLine(line,cast_to=float)
					self._nodes.append((x,y))
				elif state == 1: # read a link
					a,b = self._splitLine(line,cast_to=int)
					self._links.append((a,b))

		if self._seed is not None:
			self._autogen(self._seed, self._total_nodes, self._total_links)

		self._total_nodes = len(self._nodes)
		self._total_links = len(self._links)

		self.spreadNodes()

	def newNode(self):
		for i in xrange(100): # max attempts
			retry = False
			x,y = random.random(), random.random()
			for nx,ny in self._nodes:
				dx = nx - x
				dy = ny - y
				rs = dx*dx+dy*dy
				if rs < 0.01:
					retry = True
					break
			if not retry:
				return x,y
		return None

	def newNode(self):
		best_d = 0
		best_pos = None
		successes = 0
		threshold = 0.1
		for i in xrange(100): # max attempts
			retry = False
			x,y = random.random(), random.random()

			dmin = 100000

			for nx,ny in self._nodes:
				dx = nx - x
				dy = ny - y
				rs = dx*dx+dy*dy
				dmin = min(dmin, rs)
				if rs < 0.01:
					retry = True
					break

			if retry:
				continue

			if best_d < dmin:
				best_d = dmin
				best_pos = (x,y)
				successes += 1
				print "BEST D ",best_d, threshold
				if successes == 5 or best_d > threshold:
					break

		return best_pos

	def newLink(self):
		n = len(self._nodes)-1
		for i in xrange(100): # max attempts
			a,b = random.randint(0,n), random.randint(0,n)
			if a == b: continue
			if (a,b) in self._links: continue
			return (a,b)
		return None

	def _autogen(self, seed, total_nodes, total_links):
		random.seed(seed)

		# do we need to generate the ndoes?
		n = total_nodes - len(self._nodes)
		if n>0:
			for i in xrange(n):
				node = self.newNode()
				if node:
					self._nodes.append(node)

		if seed is not None:
			random.seed(seed * 10)

		n = total_links - len(self._links)
		if n>0:
			for i in xrange(n):
				link = self.newLink()
				if link:
					self._links.append(link)

		self._total_nodes = len(self._nodes)
		self._total_links = len(self._links)



	def spreadNodes(self):
		v = N.array(self._nodes)
		max = N.max(v,0)
		min = N.min(v,0)


		self._nodes = (v - [min]) / [max-min]
