"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from scene import Scene
from sdl2 import *
import sdl2.ext
from config import C
from cell import Cell
from trail import SplineTrail
from resources import R
import random
import numpy as N
from sprite import StaticSprite, Sprite,AnimatedSprite
from position import Position
import math
import itertools
from entity import Entity
from particles import Particles
from text import Text
from powerdial import PowerDial
from turngauge import TurnGauge
from score import Score
from timer import Timer
from player import Player, USER, ENEMY
from button import Button
from level import Level


def arcPoints(p1, p2):
	v1 = N.array(p1,dtype=float)
	v2 = N.array(p2,dtype=float)

	dir = v2-v1
	perp = N.array((dir[1], -dir[0]),dtype=float)

	vc1 = dir*0.2 + v1 + perp*0.1
	vc2 = dir*0.2 + v1 - perp*0.5
	vc3 = dir*0.8 + v1 + perp*0.1

	return (v1,vc1,vc3,v2)
	return (v1,vc1,vc2,vc3,v2)

def pick(from_seq, n):
	indices = N.arange(len(from_seq))
	N.random.shuffle(indices)
	indices = indices[:n]
	return N.take(from_seq, indices, 0)




class MainScene(Scene):

	def __init__(self, renderer, is_random):
		Scene.__init__(self, renderer)
		self._is_random = is_random
		self._score = Score(self,(C.resolution[0] / 2, 10), 0)

		self._init()

	def _init(self):
		Scene._init(self)
		self._links_from={} # cell->[cell, cell...]
		self._links_to={} # cell->[cell, cell...]
		self._link_trails={} # (cell,cell) -> trail
		self._trails_cells={} # trail -> (cell,cell)

		# when a link is to be destroyed (but the animation is running),
		# the cell pair (cell1, cell2) is passed here to mark it as invalid
		self._invalid_links = set()
		# the same as _invalid_links but for individual cells who are
		# being destroyed
		self._invalid_cells = set()
		self._game_over = False
		self._finishing = False

		self._clear_link_sound = R.sound("clear_link.wav")
		self._place_link_sound = R.sound("place_link.wav")
		self._power1_sound = R.sound("NFF-good-tip-low.wav")
		self._power2_sound = R.sound("NFF-good-tip-high.wav")
		self._game_over_die_sound = R.sound("game_over_die.wav")
		self._game_over_sound = R.sound("game_over.wav")
		self._level_done = R.sound("level_done.wav")
		self._power_active = R.sound("power_active.wav")

		self._bg = None
		self._players = None

		self.stopMusic()



	def setLevel(self, level, lev_num=None, tot_levels=None):
		self._level = level
		self._level_num = lev_num
		self._tot_levels = tot_levels

		self.clearEntities()

		level_area = N.array([[50,150],[C.resolution[0]-50, C.resolution[1]-80]], dtype=float)

		level_size = level_area[1] - level_area[0]

		cells = []

		for i, node_pos in enumerate(level.nodes):
			pos = (level_size * [node_pos] + level_area[0])[0]
			cell = self.newCell(pos, random.random()*0.5 + 0.5)
			cells.append(cell)

		for link_from, link_to in level.links:
			self.connectCells(cells[link_from], cells[link_to])

		cell = cells[level.good_blob_start_node]
		blob = self.newBlob("good-blob",pos = cell.position, scale=0.6, tint=(0.5,1.0,0.5))
		power = PowerDial(self,(40,60))
		turn_timer = TurnGauge(self,(10,10), tint = (0.8,1.0,0.8),
		                       time = level.user_time
		                       )

		user = Player(self, USER, blob, cell, power, turn_timer)

		cell = cells[level.bad_blob_start_node]
		blob = self.newBlob("bad-blob",pos = cell.position, scale=0.7, frames_per_second=35, tint=(1.0,0.5,0.5))
		power = PowerDial(self,(C.resolution[0] - 40 - power.width,60))
		enemy = Player(self, ENEMY, blob, cell, power, None)
		turn_timer = TurnGauge(self,(C.resolution[0] - 10 - turn_timer.width,10),
		                       time = level.enemy_time,
		                       on_full = self._onAutoMovePlayer,
		                       cb_data = (enemy,True), tint = (1.0,0.8,0.8))
		enemy.turn_timer = turn_timer


		self._bg = self.newStaticSprite((0,0), level.bg_image or "bg1.jpg",1.0,False,-100)

		self._score = Score(self,(C.resolution[0] / 2, 10), 0)


		self.playMusic(level.music or C.default_music)

		self._players = {USER: user, ENEMY: enemy}

		self._buttons = []

		if self._is_random:
			self.addButton(0)

		self.addButton(1)
		self.addButton(2)
		self.addButton(3)

		if lev_num or tot_levels:
			if tot_levels is None:
				level = "%s"%lev_num
			else:
				if lev_num is None:
					level = "*/%s"%tot_levels
				else:
					level = "%s/%s"%(lev_num+1,tot_levels)
			Text(self,R.font(C.main_font),(C.resolution[0]/2, 60), level, color=(1.0,1.0,0.5),center_x=True)



		self._score.setValue(C.initial_score)

	def addButton(self, n):
		bg = R.texture("button_bg2.png", blend_mode=SDL_BLENDMODE_BLEND, tint=(1,1,1))
		text = ["RANDOMIZE","RETURN","REPLAY","QUIT","NEXT"]
		pos = [(300,50),(C.resolution[0] - 380,50),(380,50),(C.resolution[0] - 300,50),(300,50)]


		button = Button(self, Text(None, C.main_font, text=text[n],size=10, color=(0,0,0)).toTexture(),pos[n],
		                background=bg, padding=0,min_size=(70,70), on_click_cb=self._onButtonClick, cb_data=n)

		self._buttons.append(button)

		if n == 4:
			 Particles(
				self,
				R.texture("sparkle.png", tint=(0.2,0.9,0.9), blend_mode=SDL_BLENDMODE_ADD),
				pos = pos[4],
				scale = 0.5,
				max_particles = 100,
				max_particle_life = 300,
				creation_rate = 100,
				continuous = False,
				full = True,
				z = 0
			)


	def _onButtonClick(self, button, data):
		print "Clicked: ",data
		if data == 0: # randomize
			self.stopMusic()
			self._init()
			level = Level()
			level.fullRandom()
			self.setLevel(level)
		elif data == 1: # return
			self.stopMusic(1000)
			self.switchScene(Scene.MENU_SCENE)
			return
		elif data == 2: # replay
			self.stopMusic()
			self._init()
			self.setLevel(self._level, self._level_num, self._tot_levels)
			return
		elif data == 3: # quit
			self.stopMusic()
			self.quitGame()
			return
		elif data == 4: # next
			self.stopMusic()
			self.quitScene() # quit scene to get the scene loaded again with the next level



	def _onAutoMovePlayer(self, _, data):
		if self._game_over: return
		player, first = data
		if first: # add a random time, so the enemy doesn't play instantly
			Timer(self, random.random()*1.0+0.5, on_done=self._onAutoMovePlayer, cb_data=(player, False))
			return
		else:
			player.play()


#-------- Graph handling (low level, don't call these directly)

	def addLink(self, cell1, cell2, trail):
		cell_pair = (cell1, cell2)

		self._invalid_links.discard(cell_pair)

		if cell_pair in self._link_trails: return

		try:
			self._links_from[cell1].append(cell2)
		except:
			self._links_from[cell1]=[cell2]

		try:
			self._links_to[cell2].append(cell1)
		except:
			self._links_to[cell2]=[cell1]


		self._link_trails[cell_pair] = trail
		self._trails_cells[trail] = cell_pair



	def thrashLink(self, cell1, cell2):
		cell_pair = (cell1, cell2)

		if cell_pair in self._invalid_links: return

		try:
			trail = self._link_trails[cell_pair]
		except:
			return None

		self._invalid_links.add(cell_pair)

		return trail


	def removeLink(self, cell1, cell2):
		cell_pair = (cell1, cell2)

		self._invalid_links.discard(cell_pair)

		if cell_pair not in self._link_trails: return None

		try:
			self._links_from[cell1].remove(cell2)
		except:
			pass

		try:
			self._links_to[cell2].remove(cell1)
		except:
			pass

		trail = self._link_trails[cell_pair]


		del self._link_trails[cell_pair]
		del self._trails_cells[trail]

		return trail


	def removeCell(self, cell):
		self._invalid_cells.discard(cell)

#-------- Graph handling (low level still, but safe to call)

	def getLink(self, cell1, cell2):
		cell_pair = (cell1, cell2)

		if cell_pair in self._invalid_links: return None

		try:
			return self._link_trails[cell_pair]
		except:
			return None

	def reachableCells(self, cell):
		try:
			return filter(lambda c: c not in self._invalid_cells and (cell,c) not in self._invalid_links, self._links_from[cell])
		except:
			return []

	def incomingCells(self, cell):
		try:
			return filter(lambda c: c not in self._invalid_cells and (c,cell) not in self._invalid_links, self._links_to[cell])
		except:
			return []


	def allCells(self):
		return filter(lambda c: c not in self._invalid_cells,self._entities_by_type[Entity.CELL_TYPE])


#-------- Graph handling (high level)

	def connectCells(self, cell1, cell2, animate = 0, on_done=None, cb_data=None):
		trail = self.newTrail()
		trail.setRoute(arcPoints(cell1.position, cell2.position))

		self.addLink(cell1, cell2, trail)

		if animate is True:
			animate = trail.animate_time

		if animate>0:
			self._place_link_sound.play()
			trail.setSpan(0,0)
			trail.animateSpan(0,1, time=animate, on_done=on_done, cb_data = cb_data)

		return trail

	def disconnectCells(self, cell1, cell2, animate = 0, reverse_anim = False, on_done=None, cb_data=None):
		# we thrash it now, and remove it when the trail gets destroyed
		trail = self.thrashLink(cell1, cell2)

		if not trail: return


		if on_done:
			on_done_fn = lambda trail, data: [trail.destroy(), on_done(trail, data)]
		else:
			on_done_fn = trail.destroy

		if animate is True:
			animate = trail.animate_time

		if animate>0:
			self._clear_link_sound.play()
			if reverse_anim:
				trail.animateSpan(0,0, time=animate, on_done=on_done_fn, cb_data = cb_data)
			else:
				trail.animateSpan(1,1, time=animate, on_done=on_done_fn, cb_data = cb_data)
		else:
			trail.destroy()


	def destroyCell(self, cell, animate=False):
		if cell in self._invalid_cells: return

		self._invalid_cells.add(cell)

		for cell2 in self._links_from[cell]:
			trail = self.getLink(cell, cell2)
			self.disconnectCells(cell, cell2, True, reverse_anim= False)

		for cell2 in self._links_to[cell]:
			trail = self.getLink(cell2, cell)
			self.disconnectCells(cell2, cell, True, reverse_anim = True)

		cell.destroy()


	# this gets called by the entity when an entity's get destroyed
	def _unregisterEntity(self, entity):
		# depending on the entity type is what we need to disconnect from the graph
		et = entity.entityType()

		if et == Entity.SPLINE_TRAIL_TYPE:
				print "Removing link"
				self.removeLink(*self._trails_cells[entity])

		elif et == Entity.CELL_TYPE:
				print "Removing cell"
				self.removeCell(entity)

		# remove the rest of the references to the entity
		Scene._unregisterEntity(self, entity)

#-------- Entity creation


	def newStaticSprite(self, pos, texture, scale = 1.0,centered = True, z=0):
		sprite = StaticSprite(self,texture, pos, scale, centered, z)
		return sprite

	def newCell(self, pos, *args, **kargs):
		cell = Cell(self, pos, *args,  listener=self,**kargs)
		return cell

	def newTrail(self, texture="drop.png"):
		path = SplineTrail(self, texture = texture)
		return path

	def newBlob(self, animation_name, pos, tint=None, *args, **kargs):
		blob = AnimatedSprite(self, R.animation(animation_name, SDL_BLENDMODE_ADD, tint=tint), pos, *args, **kargs)
		return blob

	def newParticles(self, texture, pos, *args, **kargs):
		particles = Particles(self, texture, pos, *args, **kargs)
		return particles

	def newText(self, font, pos, text, color=None, center_x = False, center_y = False, size = 30):
		text = Text(self, font, pos, text, color, center_x, center_y, size=size)
		return text


#-------- Callbacks

	def _onClick(self, what, x, y):
		print "Click", what.__class__
		if isinstance(what, Cell):
				user = self.getPlayer(USER)
				if user.is_animating or not user.turn_timer.isFull():
					return

				user = self.getPlayer(USER)

				cell1, cell2 = user.current_cell, what

				r = False

				if user.power.isActive():
					# depending on the level, is what we do...
					lev = user.power.getLevel()

					user.power.clearLevel()

					if lev == 2:
						r = self.moveLinks(cell1, cell2, 1, USER)
					elif lev == 3:
						r = self.moveLinks(cell1, cell2, 3, USER)
				else:
					r = self.moveLinks(cell1, cell2, 0, USER)

				if r: user.turn_timer.reset()

#--------

	def moveLinks(self, cell1, cell2, type, who):
		follow_trail = None
		link = self.getLink(cell1, cell2)
		if link:
			follow_trail = link
			if type == 0:
				self.disconnectCells(cell1,cell2, True,
				                     on_done = self._onLinkConsumed, cb_data=who)
			elif type == 1:
				pass
			elif type == 2:
				self.disconnectCells(cell1,cell2, True,
				                     on_done = self._onAddLink, cb_data=(cell2, cell1, who, True))
			elif type == 3 or type == 4:
				Timer(self, link.animate_time,
				    on_done = self._onAddLink, cb_data=(cell2, cell1, who, False))
		else:
			if type == 0:
				return False # can't do this
			elif type == 1 or type == 2: # create a temporary trail, and destroy it when done
				follow_trail = self.connectCells(cell1, cell2, True,
				                                 on_done=self._onDestroyLink, cb_data=(cell1, cell2, who, True))
			elif type == 3:
				follow_trail = self.connectCells(cell1, cell2, True,
				                                 on_done=self._onLinkAdded, cb_data=who)
			elif type == 4:
				follow_trail = self.connectCells(cell1, cell2, True,
				                                 on_done=self._onLinkReverse, cb_data=(cell1, cell2, who, True))
			elif type == 5:
				follow_trail = self.connectCells(cell1, cell2, True,
				                                 on_done=self._onAddLink, cb_data=(cell2, cell1, who, False))


		if follow_trail:
			self.beginAnimating(who)
			self.setBlobCell(who, cell2)
			self.getBlob(who).followTrail(follow_trail,
										  on_done = self._onDoneAnimating,
										  cb_data = who,
										  time = follow_trail.animate_time+0.2)
		return True
#----------- moveLinks callbacks


	def _onLinkAdded(self, trail, who):
		if who == USER:
			self.updateScore(trail.length * C.user_link_add_points)
		else:
			self.updateScore(trail.length * C.enemy_link_add_points)

	def _onLinkConsumed(self, trail, who):
		player = self.getPlayer(who)
		player.power.incLevel()
		if who == USER:
			lev = player.power.getLevel()
			if lev == 2:
				self._power1_sound.play()
			elif lev == 3:
				self._power2_sound.play()

			self.updateScore(trail.length * C.user_link_remove_points)
		else:
			self.updateScore(trail.length * C.enemy_link_remove_points)

	def _onAddLink(self, trail, params):
		cell1, cell2, who, link_was_consumed = params
		if not self.getLink(cell1, cell2): #only add a link if not one there already
			self.connectCells(cell1, cell2, True, on_done=self._onLinkAdded, cb_data=who)
		else:
			if link_was_consumed:# couldn't create link, so a link was actually consumed
				self._onLinkConsumed(trail, who)


	def _onDestroyLink(self, _, params):
		cell1, cell2, who, link_was_added = params
		if link_was_added: # zero sum
			self.disconnectCells(cell1, cell2, True)
		else:
			self.disconnectCells(cell1, cell2, True, on_done = self._onLinkConsumed, cb_data=who)

	def _onLinkReverse(self, trail, params):
		cell1, cell2, who, link_was_added = params
		self.disconnectCells(cell1, cell2, True)
		if not self.getLink(cell2, cell1): #only add a link if not one there already
			self.connectCells(cell2, cell1, True)
		elif not link_was_added:
			self._onLinkConsumed(trail, who)

	def _onDoneAnimating(self, _, who):
		self.endAnimating(who)
		if who == USER:
			self.getPlayer(ENEMY).play()
		self.checkGameStatus()

#--------

	def getPlayer(self, who):
		return self._players[who]

	def beginAnimating(self, who):
		print "Begin animating: ", self._players[who]
		self._players[who].is_animating = True

	def endAnimating(self, who):
		print "End animating:", self._players[who]
		self._players[who].is_animating = False

	def isAnimating(self, who):
		return self._players[who].is_animating

	def getBlob(self, who):
		return self._players[who].blob

	def setBlobCell(self, who, cell):
		self._players[who].current_cell = cell

	def getBlobCell(self, who, cell):
		return self._players[who].current_cell
#--------


	def computeScore(self):
		tot_cells = len(self._entities_by_type[Entity.CELL_TYPE])
		tot_links = len(self._link_trails)
		score =  tot_cells * 10000 + tot_links * 5000
		#print "Score: cells: %s x 10000 + Links: %s x 5000 = %s"%(tot_cells, tot_links, score)
		return score


	def updateScore(self, delta):
		if self._game_over: return
		self._score.delta(delta)


	def checkGameStatus(self):
		if self._game_over: return


		user = self.getPlayer(USER)
		enemy = self.getPlayer(ENEMY)

		# user and enemy landing in same spot -> instant game over
		if not user.is_animating and not enemy.is_animating and user.current_cell == enemy.current_cell:
			self.gameOver(True)
			return

		# user can't move anymore -> play plan (speed things up)
		if user.isLocked():
			print "**** USER LOCKED ****"
			enemy.setKillMode(True)
			enemy.turn_timer.fullSpeed()
			self._finishing = True


		# can enemy move??
		if enemy.isLocked():
			print "**** ENEMY LOCKED ****"
			if user.isLocked(2):
				self.levelOver()

		enemy.play()

		if not self._finishing and not (user.is_animating or enemy.is_animating):
			if self._score.isZero():
				self.gameOver(False)


	def gameOver(self, bad_end):
		if self._game_over: return
		self._game_over = True
		user = self.getPlayer(USER)
		if bad_end:
			user.blob.extinguish()
			self._game_over_die_sound.play()
			Particles(self, R.texture("sparkle.png", SDL_BLENDMODE_ADD,(1.0,0.5,0.5)), pos = user.current_cell.position,
			          max_particles = 300,
			          max_particle_life = 3000,
			          creation_rate = 100,
			          scatter = 7.0,
			          continuous = False,
			          accel=(0,0.00001)
			          )

			Timer(self, 1, on_done=self.onShowGameOver)
		else:
			self.onShowGameOver()

		self.stopMusic(3000)


	def levelOver(self):
		if self._game_over: return
		self._level_done.play()
		self._game_over = True
		self.showBanner("Level\nComplete!", color=(0.3,0.9,0.3))
		self.stopMusic(3000)
		self.addButton(4)


	def onShowGameOver(self, *args): # we don't care about the args
		self._game_over_sound.play()
		blend_mode = SDL_BLENDMODE_BLEND
		self.showBanner("Game\nOver", color=(0.98,0.25,0.25))


	def showBanner(self, text, color = (1,1,1), size=150, animate=True, blend_mode = SDL_BLENDMODE_BLEND, font="Semi-Casual.ttf"):
		lines = filter(lambda s:len(s)>0, map(lambda s:s.strip(), text.split()))
		n = len(lines)
		h = n*size
		y = (C.resolution[1]-h)/2
		x = C.resolution[0]/2
		px = C.resolution[0]
		dy = float(h)/n

		for line in lines:
			if animate:
				pos = (x + px,y)
			else:
				pos = (x,y)

			text = Text(self, C.main_font, pos, line, color, True, True, z=1000, size=size, blend_mode = blend_mode)

			if animate:
				id = self._springs.addMass(pos, target=(int(x)+.5,y), spring_k=7, damping=5.0)
				self._springs.manage(id, text)

			y += dy
			px = -px


#--------

	def processEvents(self):
		for event in sdl2.ext.get_events():
			event_type = event.type

			if event_type == SDL_QUIT:
				self.quitGame()
				break

			if not self._game_over and not self._finishing:
				if event_type == SDL_KEYDOWN:
					v = event.key.keysym.sym
					if v == SDLK_RETURN or v == SDLK_SPACE:
						power = self.getPlayer(USER).power
						if power.getLevel()>1:
							power.activate()
							self._power_active.play()

				for li in self._listeners:
					if li.processEvent(event): break
			else:
				for li in self._buttons:
					if li.processEvent(event): break

