"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from entity import Entity
from verlet import Verlet
from sdl2 import *
import numpy as N
import random
from resources import R

class Particles(Entity):
	def __init__(self, scene, texture, pos, scale=1,
	             max_particles=100, max_particle_life=1000, creation_rate=100,
	             accel = (0,0.000001), speed = None, scatter = 5.0, custom_generator=None,
	             continuous = True, damping = 0.001,
	             rot_speed_min = -1.0, rot_speed_max = 1.0, full =False,z=0):
		Entity.__init__(self, scene,z)
		self._spawn_max = -1 if continuous else max_particles
		self._max_particle_life = max_particle_life
		self._creation_rate = creation_rate * 0.001 # particles per second
		self._head = 0
		self._tail = 0
		self._total_particles = 0
		self._accel = accel or (0.0,0.0)
		self._spawned = 0
		self._shutdown = 0
		self._max_particles = int(min(max_particles, max_particle_life*self._creation_rate))

		if self._max_particles > 100:
			print "!!! Max particles capped to 100"
			self._max_particles = 100

		self._acum_t = self._max_particles / self._creation_rate if full else 0
		self._rot_speed = N.random.rand(max_particles) * (rot_speed_max - rot_speed_min) + rot_speed_min

		self._pos = pos
		self._scale = scale
		self._rect = SDL_Rect(0,0,1,1)

		self._generator = self.defaultGenerator if not custom_generator else custom_generator

		try:
			w = texture.width
		except:
			texture = R.texture(texture, SDL_BLENDMODE_ADD, (0.1,0.1,0.1))

		self._texture = texture
		self._renderer =texture.renderer
		self._orig_w = texture.width * scale
		self._orig_h = texture.height * scale
		self._scatter = scatter
		self._speed = speed or (0.0,0.0)

		assert(self._texture)
		assert(self._renderer)

		# a rather intrusive way of using the Verlet class... but there's no time to do it better
		self._verlet = Verlet(scene.targetFps(),initial_size=self._max_particles)
		self._verlet._tot_items = self._max_particles
		self._verlet._s[:,:] = [self._verlet.dampingToConservation(damping), 0]
		self._verlet._p[:,3,:] = accel or [0,0]

		print "Maximum simultaneous particles: ",self._max_particles



	def entityType(self):
		return Entity.PARTICLES_TYPE

	def categories(self):
		from scene import Scene
		return (Scene.RENDERABLE,)

	def moveTo(self, pos):
		self._pos = pos

	def defaultGenerator(self, pos):
		while True:
			dx = 2.*random.random()-1.
			dy = 2.*random.random()-1.

			r = dx*dx+dy*dy

			if r < 1.0:
				break
		s = self._speed
		speed =N.array([dx*self._scatter+s[0],dy*self._scatter+s[1]])
		return pos-speed, pos


	def shutdown(self):
		self._shutdown = True

	def update(self, t, dt):
		self._acum_t += dt
		new_particles = min(int(self._acum_t * self._creation_rate), self._max_particles)
		self._acum_t -= new_particles / self._creation_rate

		if new_particles>0:
			max_p = self._max_particles
			if self._shutdown:
				if self._total_particles == 0:
					return lambda:self.destroy()

				n = min(new_particles, self._total_particles)
				self._tail += n
				while self._tail >= max_p: self._tail -= max_p
				self._total_particles -= n
			else:
				pos = self._pos
				for i in xrange(new_particles):
					loc = self._head

					next_head = loc+1
					if next_head == max_p:
						next_head = 0
					self._total_particles += 1
					self._spawned += 1

					if self._total_particles >= max_p:
						self._total_particles -= 1
						self._tail = self._tail+1
						if self._tail == max_p:
							self._tail = 0
					self._head = next_head

					self._verlet._p[loc,0:2,:] = self._generator(self._pos)
					self._verlet._p[loc,2] = self._accel

					if self._spawned == self._spawn_max:
						self._shutdown = True

		self._verlet.step()

		return None

	def paint(self):
		if self._total_particles ==0: return
		t = float(self._total_particles) / self._max_particles
		dt = 1.0 / self._total_particles
		tw = self._orig_w
		th = self._orig_h
		p = self._verlet._p[:,1]
		a = self._rot_speed
		renderer = self._renderer
		texture = self._texture.texture
		rect = self._rect

		n = max(self._head, self._total_particles)
		for i in xrange(self._tail, n):
			if t < 0.1: # fade in/out of particles
				s = max(t*10.0,0.0)
			elif t > 0.9:
				s = max((1.0 - t) * 10.0,0.0)
			else:
				s = 1.0

			w = tw * s
			h = th * s
			rect.w = int(w)
			rect.h = int(h)
			rect.x = int(p[i,0]-w*0.5)
			rect.y = int(p[i,1]-h*0.5)
			SDL_RenderCopyEx(renderer, texture, None, rect, a[i] * t, None, SDL_FLIP_NONE)

			t -= dt

		# we're using a cyclic buffer, so we might need to paint it in two parts

		if n != self._head:
			for i in xrange(self._head):
				if t < 0.1: # fade in/out of particles
					s = max(t*10.0,0.0)
				elif t > 0.9:
					s = max((1.0 - t) * 10.0,0.0)
				else:
					s = 1.0

				w = tw * s
				h = th * s
				rect.w = int(w)
				rect.h = int(h)
				rect.x = int(p[i,0]-w*0.5)
				rect.y = int(p[i,1]-h*0.5)
				SDL_RenderCopyEx(renderer, texture, None, rect, a[i] * t, None, SDL_FLIP_NONE)
				t -= dt
