"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from sdl2 import *
import sdl2.ext
from config import C
from resources import R
from verlet import Verlet

class Scene(object):

	RENDERABLE=1
	STATIC =2
	LISTENER = 3
	CONTAINER = 4

	MENU_SCENE = 1
	MAIN_SCENE = 2
	HELP_SCENE = 3
	CREDITS_SCENE = 4

	def __init__(self, renderer):
		self._change_scene = 0 # 0 = still on scene, False = exit game, True = exit scene, >0 = change to a specific scene
		self._change_scene_params = None
		self._renderer = renderer
		self._bgm = None
		self._init()

	@property
	def renderer(self):
		return self._renderer

	def processEvents(self):
		for event in sdl2.ext.get_events():
			event_type = event.type

			if event_type == SDL_QUIT:
				self.quitGame()
				break

			for li in self._listeners:
				if li.processEvent(event): break


	def switchScene(self, number, params=None):
		self._change_scene  = number
		self._change_scene_params = params

	def quitGame(self, params=None):
		self._change_scene = False
		self._change_scene_params = params


	def quitScene(self, params=None):
		self._change_scene = True
		self._change_scene_params = params


	def update(self, t, dt):
		self._masses.step(False)
		self._springs.step(True)

		self._masses.signalProxies()
		self._springs.signalProxies()

		to_cleanup = []
		for e in self._updateables:
			to_do = e.update(t, dt)
			if to_do is not None:
				to_cleanup.append(to_do)

		if to_cleanup:
			#print "To cleanup:", len(to_cleanup)
			for to_do in to_cleanup:
				to_do()

	def render(self):
		renderer = self.renderer

		SDL_SetRenderDrawColor(renderer, 0, 0, 0, 255)
		SDL_RenderClear(renderer)

		for e in self._renderables:
			e.paint()

		SDL_RenderPresent(renderer)


	def getNewEntityId(self):
		self._max_entity_id += 1
		return self._max_entity_id

	# this shouldn't be called manually. The entity calls this
	def _registerEntity(self, entity):
		et = entity.entityType()
		if et is None:
			print "Can't register a pure Entity"
			return

		categories = entity.categories()

		if not isinstance(categories, (list, tuple)):
			categories = (categories,)

		self._entities[entity.entity_id] = entity

		try:
			self._entities_by_type[et].append(entity)
		except:
			self._entities_by_type[et]=[entity]


		if Scene.RENDERABLE in categories:
			self._renderables.append(entity)
			self._renderables.sort(key=lambda a:a.z_order)
			#print "---"
			#for r in self._renderables:
			#print r.__class__, r.z_order

		if Scene.STATIC not in categories:
			self._updateables.add(entity)

		if Scene.LISTENER in categories:
			self._listeners.add(entity)


	# this shouldn't be called manually. The entity calls this
	def _unregisterEntity(self, entity):
		et = entity.entityType()
		if et is None:
			print "Can't unregister a pure Entity"
			return

		categories = entity.categories()

		if not isinstance(categories, (list, tuple)):
			categories = (categories,)

		del self._entities[entity.entity_id]

		self._entities_by_type[et].remove(entity)

		if Scene.RENDERABLE in categories:
			self._renderables.remove(entity)

		if Scene.LISTENER in categories:
			self._listeners.remove(entity)

		if Scene.STATIC not in categories:
			self._updateables.remove(entity)


	def _init(self):
		self._springs = Verlet(1.0/self.targetFps())
		self._masses = Verlet(1.0/self.targetFps())
		self._renderables= []
		self._updateables = set()
		self._listeners = set()
		self._entities = {}
		self._max_entity_id = 0
		self._entities_by_type= {}

	def entities(self, entity_type):
		try:
			return self._entities_by_type[entity_type]
		except:
			return []

	def clearEntities(self):
		for e in self._entities.values():
			e.destroy()
		self._init()



	def _perSecond(self, dt, t):
		fps = float(self._current_frame - self._last_fps_frame) / dt * 1000.0
		print "FPS: ",fps
		self._last_fps_frame = self._current_frame



	def loop(self):
		cap_frames = True
		frame_t = 1000.0 / C._framerate
		last_t = SDL_GetTicks() - 1
		target_t = last_s = last_t
		self._current_frame = 0
		self._last_fps_frame = 0
		drop = False
		while self._change_scene is 0:
			self._current_frame += 1

			t = float(SDL_GetTicks())
			dt = t - last_t
			# ----

			self.processEvents()
			self.update(t, dt)
			if not drop:
				self.render()

			drop = False

			# timer stuff
			last_t = t
			if t - last_s > 1000:
				if C.print_fps:
					self._perSecond(t-last_s, t)
				last_s = t

			if cap_frames:
				target_t += frame_t
				if t < target_t:
					SDL_Delay(int(target_t - t))
				else:
					target_t = t # below FPS
					drop = True
		self.stopMusic()
		return self._change_scene, self._change_scene_params


	def targetFps(self):
		return C._framerate


	def playMusic(self, music, fade_in_ms = 1000):
		if self._bgm:
			self._bgm.stop()

		if C.play_music:
			self._bgm = R.music(music)
			self._bgm.play(fade_in_ms=fade_in_ms)


	def stopMusic(self, fade_out_ms = 0):
		if self._bgm:
			a = self._bgm
			# use the lambda just to keep a reference to the self._bgm instance
			self._bgm.stop(fade_out_ms=fade_out_ms, on_stop=lambda:[a])
		self._bgm = None

#----------------------------------

