"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from sdl2 import *
from entity import Entity

from utils import smooth
from resources import R
from animation import Animation

class StaticSprite(Entity):
	def __init__(self, scene, src, pos, scale = 1.0, centered = True, z=0):
		Entity.__init__(self, scene, z)
		self._centered = centered

		if isinstance(src, basestring):
			src = R.texture(src)
		self._src = src
		self._extinguish = 1.0
		self._extinguish_speed = 0.0
		try:
			self._orig_w = src.width
			self._orig_h = src.height
		except:
			self._orig_w = src.contents.w
			self._orig_h = src.contents.h

		self._pos = pos
		self._scale_x = scale
		self._scale_y = scale
		self._rect = SDL_Rect(0,0,1,1)
		self._renderer = R.renderer
		self.scale(scale)
		self.moveTo(pos)

	def entityType(self):
		return Entity.STATIC_SPRITE_TYPE

	def categories(self):
		from scene import Scene
		return (Scene.RENDERABLE, Scene.STATIC)

	def setCentered(self, centered):
		self._centered = centered
		self.scale(self._scale_x, self._scale_y)

	def scale(self, scale_x, scale_y = None):
		if scale_y is None: scale_y = scale_x

		self._scale_x= scale_x
		self._scale_y= scale_y
		self._w = self._orig_w * scale_x * self._extinguish
		self._h = self._orig_h * scale_y * self._extinguish
		self._rect.w = int(self._w+ 0.5)
		self._rect.h = int(self._h+ 0.5)
		if self._centered:
			self._w2 = self._w * 0.5
			self._h2 = self._h * 0.5
			self._rect.x = int(self._pos[0] - self._w2+0.5)
			self._rect.y = int(self._pos[1] - self._h2+0.5)
		else:
			self._rect.x = int(self._pos[0] + 0.5)
			self._rect.y = int(self._pos[1] + 0.5)


	@property
	def width(self):
		return self._w

	@property
	def height(self):
		return self._h

	def moveTo(self, pos):
		self._pos = pos
		if self._centered:
			self._rect.x = int(self._pos[0] - self._w2+0.5)
			self._rect.y = int(self._pos[1] - self._h2+0.5)
		else:
			self._rect.x = int(self._pos[0] + 0.5)
			self._rect.y = int(self._pos[1] + 0.5)

	@property
	def position(self):
		return self._pos


	def paint(self):
		self._src.paintRect(self._rect)

	def changeTexture(self, src):
		if src == self._src:
			return

		if isinstance(src, basestring):
			src = R.texture(src)

		self._src = src
		self._orig_w = src.width
		self._orig_h = src.height

		self.scale(self._scale_x, self._scale_y) # to recompute rect


class Sprite(StaticSprite):
	def __init__(self, scene, src, pos, scale = 1.0, centered = True, z=0):
		StaticSprite.__init__(self, scene, src, pos, scale, centered, z)
		self._physics = None
		self._anim_t = 0
		self._anim_dt = None
		self._trail = None
		self._on_done_anim_fn = None
		self._to_do = []
		self._t = 0
		self._dt = 0

	def entityType(self):
		return Entity.SPRITE_TYPE

	def categories(self):
		from scene import Scene
		return (Scene.RENDERABLE,)

	def addToVerlet(self, verlet, *args, **kargs):
		id = verlet.addMass(pos=self._pos, damping=2, *args,**kargs)
		self._physics = verlet.getProxy(id, self._onChange)

	def _onChange(self):
		self.moveTo(self._physics.value()[1])

	def setTarget(self, pos):
		if self._physics is None: return
		self._physics.params()[1][1] = pos


	def followTrail(self, trail, time=2.0, on_done=None, cb_data=None):
		self._anim_t = 0
		self._anim_dt = 1.0 / (time * 1000.0)
		self._trail = trail
		self._on_done_anim_fn = on_done
		self._on_done_data = cb_data


	def extinguish(self, speed=0.5): # 2 seconds by default
		self._extinguish_speed= speed * 0.001


	def followingTrail(self):
		return self._trail is not None

	def update(self, t, dt):
		self._dt = dt
		self._t = t
		to_do = None

		if self._extinguish_speed>0:
			self._extinguish -= self._extinguish_speed * dt
			if self._extinguish <= 0:
				self._extinguish = 0
				return lambda:self.destroy()

		if self._anim_dt is not None:
			self._anim_t += self._anim_dt * dt
			if self._anim_t > 1.0:
				self._anim_t = 1.0
				self._anim_dt = None
				self.moveTo(self._trail(1)[0:2])
				self._trail = None
				if self._on_done_anim_fn:
					cb = self._on_done_anim_fn
					data = self._on_done_data
					self._on_done_anim_fn = None
					self._on_done_data = None
					to_do = lambda :cb(self, data)
				if self._physics:
					self._physics.array.fix(self._physics.id, self._pos)
			else:
				self.moveTo(self._trail(smooth(self._anim_t))[0:2])

		return to_do


class AnimatedSprite(Sprite):
	def __init__(self, scene, animation, pos, scale = 1.0, frames_per_second = 30, centered = True, z=0):
		if isinstance(animation, Animation):
			pass
		elif isinstance(animation, basestring):
			animation = R.animation(animation, SDL_BLENDMODE_ADD)
		else:
			raise Exception("Unsupported type for animation source: %s"%type(animation))

		Sprite.__init__(self, scene, animation, pos, scale, centered, z)
		self._fpms = frames_per_second / 1000.0
		self._frame = 0

	def entityType(self):
		return Entity.ANIM_SPRITE_TYPE

	def categories(self):
		from scene import Scene
		return (Scene.RENDERABLE,)

	def update(self, t, dt):
		r = Sprite.update(self, t, dt)
		self._frame = int(t * self._fpms) % self._src.frames
		return r

	def paint(self):
		self._src.paint(self._frame, self._rect)

