"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from sdl2 import *
from sdl2.sdlttf import (TTF_OpenFont,
						 TTF_RenderText_Blended,
						 TTF_GetError,
						 )
from entity import Entity
from texture import Texture
import numpy as N

class Font:
	def __init__(self, font_file, renderer, size):
		self._font = TTF_OpenFont(font_file, size)
		self._size = size
		if not self._font:
			raise Exception(TTF_GetError())


	@property
	def size(self):
		return self._size

	@property
	def font(self):
		return self._font


class Text(Entity):
	def __init__(self, scene, font, pos=(0,0), text="Meow", color=None, center_x = False, center_y = False, blend_mode=SDL_BLENDMODE_BLEND, size=30, z = 10):
		from resources import R
		self._texture = None # a plain SDL (ie not a texture.Texture)

		Entity.__init__(self, scene,z=z)

		if isinstance(font, Font):
			self._font = font
		else:
			self._font = R.font(font, size)

		self._renderer = R.renderer

		self._rect = SDL_Rect(0,0,1,1)
		self._color = SDL_Color(0,0,0)
		self._prev_color = None
		self._prev_args = None
		self._blend_mode = SDL_BLENDMODE_BLEND if blend_mode is None else blend_mode

		self._pos = pos
		self._center_x = center_x
		self._center_y = center_y

		self._fillColor(N.array(color)*255 if color else (255,255,255))
		self.setText(text)
		self.setAlignment(center_x, center_y)


	def toTexture(self, tint=None):
		t = Texture.fromRaw(self._texture, self._renderer, self._width, self._height, tint)
		self._texture = None
		self.destroy()
		return t


	def entityType(self):
		return Entity.TEXT_TYPE

	def categories(self):
		from scene import Scene
		return (Scene.RENDERABLE,Scene.STATIC)

	def setColor(self, fg_color):
		self._prev_color =  fg_color


	def _fillColor(self, color):
		color =map(int,color)
		self._prev_color = color

		self._color.r = color[0]
		self._color.g = color[1]
		self._color.b = color[2]
		try:
			self._color.a = color[3]
		except:
			self._color.a = 255


	def setText(self, text, color=None):

		if color is None:
			color = self._prev_color

		t = (text, color)

		if self._prev_args == t:
			return

		self._prev_args = t

		self._text = text
		if self._texture:
			SDL_DestroyTexture(self._texture)

		if self._prev_color != color:
			self._fillColor(color)


		surf = TTF_RenderText_Blended(self._font.font, text, self._color)

		if not surf:
			raise Exception(TTF_GetError())

		self._width = surf.contents.w
		self._height = surf.contents.h

		texture = SDL_CreateTextureFromSurface(self._renderer,surf)
		if not texture:
			print "!!! Empty texture"

		if self._blend_mode is not None:
			SDL_SetTextureBlendMode(texture, self._blend_mode)

		self._texture = texture

		SDL_FreeSurface(surf)

		self._rect.w = self._width
		self._rect.h = self._height

		self.moveTo(self._pos)

	def setAlignment(self, center_x, center_y):
		self._center_x = center_x
		self._center_y = center_y
		self.moveTo(self._pos)

	@property
	def width(self):
		return self._width

	@property
	def height(self):
		return self._height

	def moveTo(self, pos):
		self._pos = pos
		if self._center_x:
			self._rect.x = int(pos[0] - self._width * 0.5)
		else:
			self._rect.x = int(pos[0])

		if self._center_y:
			self._rect.y = int(pos[1] - self._height * 0.5)
		else:
			self._rect.y = int(pos[1])


	def paint(self):
		SDL_RenderCopy(self._renderer, self._texture, None, self._rect)


	def __del__(self):
		if self._texture is not None:
			SDL_DestroyTexture(self._texture)

