"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""


from sdl2 import *
import sdl2.render
import ctypes

class Texture:

	# actually suface can be another texture, in which case the texture is cloned
	def __init__(self, surface, renderer, blend_mode=SDL_BLENDMODE_BLEND, tint=None):
		if not surface: print "!!! Empty surface"
		if ("SDL_Texture" in str(type(surface))): # from raw
			self._texture = surface
			self._renderer = renderer
			self._width = 10 # will be set later on
			self._height = 10
		elif isinstance(surface, Texture): # clone
			self._texture = surface._texture
			self._renderer = surface._renderer
			if not tint:
				tint = surface._tint
			self._tint = tint or surface._tint
			self._width = surface._width
			self._height = surface._height
		else:
			self._texture = SDL_CreateTextureFromSurface(renderer, surface)
			self._renderer = renderer
			self._width = surface.w
			self._height = surface.h

			if blend_mode is not None:
				SDL_SetTextureBlendMode(self._texture, blend_mode)

		self._old_tint = None
		self._tint = None

		self.setTint(tint)

		if self._tint is not None:
			SDL_SetTextureColorMod(self._texture, self._tint[0], self._tint[1], self._tint[2])


		if not self._texture: print "!!! Empty texture"

	def setTint(self, tint):
		if tint == self._old_tint:
			return

		if tint is not None:
			self._tint = (int(tint[0]*255.0), int(tint[1]*255.0), int(tint[2]*255.0))
		else:
			self._tint = (254,254,254)

		self._old_tint = tint

	@classmethod
	def fromRaw(cls, raw_texture, renderer, w, h, tint=None):
		assert ("SDL_Texture" in str(type(raw_texture)))
		t = Texture(raw_texture, renderer, tint)
		t._width = w
		t._height = h
		return t


	def clone(self, tint=None):
		return Texture(self, None, tint=tint)


	def paint(self, pos, scale = None):
		if self._tint is not None:
			SDL_SetTextureColorMod(self._texture, self._tint[0], self._tint[1], self._tint[2])
		if scale is not None:
			try:
				sx,sy = scale
			except:
				sx = sy = scale

			SDL_RenderCopy(self._renderer, self._texture, None, SDL_Rect(pos.x, pos.y, int(self._width * sx), int(self._height * sy)) )
		else:
			SDL_RenderCopy(self._renderer, self._texture, None, SDL_Rect(pos.x, pos.y, self._width, self._height) )


	def paintCentered(self, pos, scale = None):
		if self._tint is not None:
			SDL_SetTextureColorMod(self._texture, self._tint[0], self._tint[1], self._tint[2])
		dx = self._width * 0.5
		dy = self._height * 0.5
		if scale is not None:
			try:
				sx,sy = scale
			except:
				sx = sy = scale

			SDL_RenderCopy(self._renderer, self._texture, None, SDL_Rect(pos.x-dx*sx, pos.y-dy*sy, self._width * sx, self._height * sy) )
		else:
			SDL_RenderCopy(self._renderer, self._texture, None, SDL_Rect(pos.x-dx, pos.y-dy, self._width, self._height) )


	def paintRect(self, dest_rect):
		if self._tint is not None:
			SDL_SetTextureColorMod(self._texture, self._tint[0], self._tint[1], self._tint[2])
		SDL_RenderCopy(self._renderer, self._texture, None, SDL_Rect(dest_rect.x, dest_rect.y, dest_rect.w, dest_rect.h) )


	def paintFill(self):
		if self._tint is not None:
			SDL_SetTextureColorMod(self._texture, self._tint[0], self._tint[1], self._tint[2])
		SDL_RenderCopy(self._renderer, self._texture, None, None)

	def paintRot(self, pos, angle):
		if self._tint is not None:
			SDL_SetTextureColorMod(self._texture, self._tint[0], self._tint[1], self._tint[2])
		SDL_RenderCopyEx(self._renderer, self._texture, None, None, angle, None, SDL_FLIP_NONE)

	@property
	def renderer(self):
		return self._renderer

	@property
	def texture(self):
		return self._texture

	@property
	def width(self):
		return self._width

	@property
	def height(self):
		return self._height


	def __del__(self):
		try:
			SDL_DestroyTexture(self._texture)
		except:
			pass
