"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

from sdl2 import *
import sdl2.ext
import numpy as N
from bspline import Bspline
import math
import random
from entity import Entity
from utils import smooth,rad2deg
from config import C

from texture import Texture


Pi2 = 2 * math.pi

class SplineTrail(Entity):
	def __init__(self, scene, texture):
		from resources import R
		Entity.__init__(self, scene)
		self._total_particles = 0
		self._anim_period = 0 # to be computed

		if isinstance(texture, basestring):
			self._texture = R.texture(texture, SDL_BLENDMODE_ADD, (0.8,0.7,0.4))
		elif isinstance(texture, Texture):
			self._texture = texture
		else:
			raise ValueError("Unkown texture type:", type(self._texture))

		self._renderer = R.renderer
		self._total_samples = 0
		self._path_length = 1.
		self._path = [(0,0),(0,0)]
		#self._sdl_points_path = None # a SDL_Point array
		self._rect = SDL_Rect(0,0,self._texture.width,self._texture.height)
		self._texture_width = self._texture.width
		self._texture_height = self._texture.height
		self._t_offset = random.random()
		self._t0 = 0
		self._t1 = 1
		self._dt0 = 0
		self._dt1 = 0
		self._t = 0
		self._on_done_anim = None


	def entityType(self):
		return Entity.SPLINE_TRAIL_TYPE

	def categories(self):
		from scene import Scene
		return (Scene.RENDERABLE,)


	def computeLength(self):
		if len(self._path)==0: return 0

		def reduceFn(prev, nxt):
			p = prev[1]
			dx = p[0] - nxt[0]
			dy = p[1] - nxt[1]
			d = math.sqrt(dx*dx+dy*dy)
			return (prev[0]+d, nxt)

		return reduce(reduceFn, self._path[1:,0:2], (0.0,self._path[0,0:2]))[0]


	def setRoute(self, points):
		tot_p = len(points)

		assert(tot_p > 1)

		total_samples = 5 * tot_p

		order = min(tot_p-1, 4)

		knot_len = tot_p + 1 + order

		knot_vector = N.zeros(knot_len, dtype = float)

		#print "Total points: ",tot_p
		#print "Order: ",order
		#print "Knot size: ",knot_len

		knot_vector[0:order+1] = 0.0
		knot_vector[-order-1:] = 1.0

		a = N.linspace(0.0,1.0,tot_p-order+1)

		knot_vector[order:-order] = smooth(a)

		#print "Knot vector: ",knot_vector

		basis = Bspline(knot_vector, order)

		self._path = N.zeros([total_samples, 4], dtype = N.float) # [x,y,angle,speed]
		self._angle = N.zeros(total_samples, dtype = N.float)
		self._speed = N.zeros(total_samples, dtype = N.float)

		s_points = N.array(points, dtype=N.float).transpose()

		angle = 0
		for i,t in enumerate(N.linspace(0.0, 1.0, total_samples)):
			self._path[i,0:2] = N.sum(basis(t*0.99999) * s_points, 1)
			dx,dy = N.sum(basis.d(t*0.99999) * s_points, 1)
			new_angle = math.atan2(dy, dx)
			if new_angle < 0: # make range 0..2 Pi
				new_angle += Pi2

			da = new_angle - (angle % Pi2)
			if da > math.pi: da -= Pi2
			if da < -math.pi: da += Pi2
			angle += da

			self._path[i,2] = angle # make the smallest angle increment, even if it goes below 0 or beyond 2 pi
			self._path[i,3] = math.sqrt(dy*dy+dx*dx)


		self._path_length = self.computeLength()
		self._total_particles = min(int(self._path_length*0.02),20)
		self._total_samples = total_samples

		#print "Total particles: ",self._total_particles
		#print "Path length: ",self._path_length

		self._anim_period = 10 * self._path_length
		#self._sdl_points_path,n = sdl2.ext.to_ctypes(map(lambda pt:SDL_Point(int(pt[0]),int(pt[1])), self._path), SDL_Point)
		#assert(n == total_samples)

	@property
	def length(self):
		return self._path_length

	def __call__(self, t):
			if t >= 1.0:
				return self._path[-1]
			elif t <= 0.0:
				return self._path[0]

			t = float(t) * (self._total_samples-1)

			n = int(t) # integer part
			ft = t - n # fractional part

			return (self._path[n+1] - self._path[n]) * ft + self._path[n]


	def setSpan(self, t0, t1):
		self._t0 = t0
		self._t1 = t1

	def animateSpan(self, t0, t1, on_done=None, cb_data = None, time = None):
		if time is None: time = self.animate_time
		time *= 1000.0 # seconds to millis
		self._dt0 = float(t0 - self._t0) / time
		self._dt1 = float(t1 - self._t1) / time
		if on_done:
			self._on_done_anim = (on_done, cb_data)
		else:
			self._on_done_anim = None


	@property
	def animate_time(self):
		return self.length * C.trail_fade_speed

	def update(self, t, dt):
		self._t = t
		self._dt = t

		if self._total_particles==0:
			return None

		done_anim = False

		if self._dt0 != 0.0:
			self._t0 += self._dt0 * dt
			if self._dt0 > 0:
				if self._t0 >= self._t1:
					self._t0 = self._t1
					self._dt0 = 0
					done_anim = True
					#print "Done 1"
			elif self._t0 <= 0.0:
				self._t0 = 0.0
				self._dt0 = 0
				done_anim = True
				#print "Done 2"

		if self._dt1 != 0.0:
			done_anim = False
			self._t1 += self._dt1 * dt
			if self._dt1 > 0 :
				if self._t1 >= 1.0:
					self._t1 = 1.0
					self._dt1 = 0
					done_anim = True
					#print "Done 3"
			elif self._t1 <= self._t0:
				self._t1 = self._t0
				self._dt1 = 0
				done_anim = True
				#print "Done 4"


		if done_anim and self._on_done_anim is not None:
			data = self._on_done_anim[1]
			callback = self._on_done_anim[0]
			fn = lambda:callback(self,data)
			self._on_done_anim = None
			return fn
		return None


	def paint(self):
		if self._total_particles==0:
			return

		renderer = self._renderer
		texture = self._texture.texture
		t = (self._t/self._anim_period+self._t_offset) % 1.0
		angle = 0
		dt = 1.0 / self._total_particles

		rect = self._rect

		tw = self._texture_width
		th = self._texture_height

		#SDL_SetRenderDrawColor(renderer, 32,0,32, SDL_ALPHA_OPAQUE)
		#SDL_SetRenderDrawBlendMode(renderer, SDL_BLENDMODE_ADD)
		#SDL_RenderDrawLines(renderer, self._sdl_points_path,self._total_samples)

		f = 4.0 / self._path_length


		for i in xrange(self._total_particles):
			if t > self._t0 and t < self._t1:
				a = (t - self._t0)
				if a > 0.1: a = 0.1
				b = (self._t1 - t)
				if b > 0.1: b = 0.1

				p = self(t)

				s = (a+b-0.1) * p[3] * f

				w = tw * s
				h = th * s
				rect.w = int(w)
				rect.h = int(h)
				rect.x, rect.y = map(int, p[0:2] - (w*0.5, h*0.5))
				SDL_RenderCopyEx(renderer, texture, None, rect, p[2]*rad2deg, None, SDL_FLIP_NONE)
			t += dt
			if t >= 1.0: t -= 1

