"""
The MIT License (MIT)

Copyright (c) 2017 Guillermo Romero Franco (AKA Gato)

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import numpy as N
from packedarr import PackedArray
import weakref


class Verlet(PackedArray):
	def __init__(self, dt, dimensions = 2, initial_size = 100):
		PackedArray.__init__(self, initial_size)
		self._dimensions = dimensions
		self._dt = dt
		self._dt_sq = dt*dt
		self._p = N.zeros([initial_size, 4,dimensions]) # position t-1, position t, acceleration, target  <--- vectors
		self._s = N.zeros([initial_size,2]) # damping, k/mass     <--- scalars
		self._q = N.zeros([initial_size,dimensions])

		self.setDefaultDamping()

		self._managed = {}


	def _resize(self, new_size):
		self._p = N.resize(self._p, [new_size, 4,self._dimensions])
		self._s = N.resize(self._s, [new_size,2])
		self._q = N.resize(self._q, [new_size, self._dimensions])


	def _moveData(self, to_loc, from_loc):
		self._p[to_loc] = self._p[from_loc]
		self._s[to_loc] = self._s[from_loc]


	def _getGetters(self):
		def val_getter(loc):
			return self._p[loc,0:2] # returns (pos(t-1), pos(t))
		def par_getter(loc):
			return (self._s[loc],self._p[loc,2:]) # returns ((energy conservation, k/mass),(accel, target))
		return val_getter, par_getter

	@property
	def dimensions(self):
		return self._dimensions

	def setDefaultDamping(self, damping=0.5):
		self._default_conservation = self.dampingToConservation(damping)

	def dampingToConservation(self, damping):
		damping = max(damping,0)
		return 2.0 / (2**damping + 1.0)

	def addMass(self, pos, speed=None, accel=None, damping=None, mass=1, spring_k=1, target=None):
		pos = N.array(pos, dtype=float)
		if speed is None:
			speed = N.zeros(self._dimensions)
		else:
			speed = N.array(speed,dtype=N.float)

		if accel is None:
			accel = N.zeros(self._dimensions)
		else:
			accel = N.array(speed,dtype=N.float)

		if damping is not None:
			damping = max(damping,0)
			conservation = 2.0 / (2**damping + 1.0)
		else:
			conservation = self._default_conservation

		loc, id = PackedArray._add(self)

		self._p[loc] = [
				pos-speed * self._dt,
				pos,
				accel,
				pos if target is None else target,
			]

		self._s[loc] = [
				conservation ** self._dt,
				float(spring_k) / float(mass)
			]

		return id

	def removeMass(self, id):
		PackedArray._remove(self,id)


	def step(self, as_springs = False):
		n = self.size

		if n == 0: return

		p = self._p[0:n]
		q = self._q[0:n]
		s = self._s[0:n]

		if as_springs:
			p[:,2] = s[:,1:2] * (p[:,3] - p[:,1])

		q[:] = p[:,1] + s[:,0:1] * (p[:,1]-p[:,0]) + p[:,2] * self._dt_sq
		p[:,0] = p[:,1]
		p[:,1] = q


		if self._managed:
			to_remove = None
			for id, data in self._managed.iteritems():
				obj, holds_id = data
				obj = obj()
				if obj is None: # object disappeared
					if to_remove is None:
						to_remove=[(id,holds_id)]
					else:
						to_remove.append((id,holds_id))
					continue
				loc = self.loc(id)
				obj.moveTo(p[loc,1])

			if to_remove:
				for id, holds_id in to_remove:
					del self._managed[id]
					if holds_id:
						self._remove(id)




	def pos(self, id, new_pos=None):
		if new_pos is None:
			return self._p[self.loc(id),1]
		else:
			p = self._p[self.loc(id)]
			delta = p[1] - p[0]
			p[1] = new_pos
			p[0] = new_pos - delta


	def target(self, id, new_target=None):
		if new_target is None:
			return self._p[self.loc(id),3]
		else:
			self._p[self.loc(id),3] = new_target


	def fix(self, id, new_target):
		loc = self.loc(id)
		self._p[loc,0] = new_target # old position
		self._p[loc,1] = new_target # new position
		self._p[loc,3] = new_target # target



	def speed(self, id, new_speed=None):
		p = self._p[self.loc(id)]
		if new_speed is None:
			return (p[1] - p[0]) / self._dt
		else:
			p = self._p[self.loc(id)]
			p[0] = p[1] - new_speed * self._dt


	def accel(self, id, new_accel=None):
		if new_accel is None:
			return self._p[self.loc(id),2]
		else:
			self._p[self.loc(id),2] = new_accel

	def iter(self):
		return self._p[:self.size]

	# obj_holds_id means the object is the owner of the id, and the
	# id in the verlet must disappear when the object disappears
	def manage(self, id, obj, obj_holds_id=True): # obj must implement moveTo
		self._managed[id] = (weakref.ref(obj), obj_holds_id)


if __name__ == "__main__":
	print("test...")

	steps = 1000
	dt = 1.0/60.0

	ver = Verlet(dt)

	ver.addMass([0,0,0], damping = 1, target=[1,2,3]) # orange
	ver.addMass([0,0,0], damping = 2, target=[1,2,3]) # blue
	ver.addMass([0,0,0], damping = 2, target=[1,2,3], spring_k = 20) # green

	values = N.zeros([steps, ver.dimensions * ver.size])

	for i in xrange(steps):
		ver.step(True)
		for j in xrange(ver.size):
			values[i,j*3:j*3+3] = ver.pos(j)

	x = N.linspace(0,steps*dt,num=steps)

	import matplotlib.pyplot as plt

	args = []
	styles=["#ff0000","#ff3333","#ff6666","#0000ff","#3333ff","#6666ff","#00ff00","#33ff33","#66ff66"]
	for i in xrange(values.shape[1]):
		args.append(x)
		args.append(values[:,i])
		args.append(styles[i])

	plt.plot(*args)
	plt.show()



